/*
 * Copyright (c) 2019 NIBIO <http://www.nibio.no/>. 
 * 
 * This file is part of SeptoriaReferenceHumidityModel.
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 * 
 */

package no.nibio.vips.model.septoriareferencehumiditymodel;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;
import no.nibio.vips.entity.ModelConfiguration;
import no.nibio.vips.entity.Result;
import no.nibio.vips.entity.ResultImpl;
import no.nibio.vips.entity.WeatherObservation;
import no.nibio.vips.model.ConfigValidationException;
import no.nibio.vips.model.Model;
import no.nibio.vips.model.ModelExcecutionException;
import no.nibio.vips.model.ModelId;
import no.nibio.vips.util.ModelUtil;
import no.nibio.vips.util.WeatherElements;
import no.nibio.vips.util.WeatherUtil;
import no.nibio.vips.util.XDate;

/**
 * @copyright 2019 <a href="http://www.nibio.no/">NIBIO</a>
 * @author Tor-Einar Skog <tor-einar.skog@nibio.no>
 */
public class SeptoriaReferenceHumidityModel implements Model {
    
    public final static ModelId MODEL_ID = new ModelId("SEPTREFHUM");
    
    private final Integer MINIMUM_HOURLY_VALUES = 23;
    
    private final Double TEMPERATURE_THRESHOLD = 8.0;
    private List<WeatherObservation> observations;
    private TimeZone timeZone;

    @Override
    public List<Result> getResult() throws ModelExcecutionException {
        
        if(this.observations == null)
        {
            throw new ModelExcecutionException("No weather data. Aborting.");
        }
        
        List<Result> retVal = new ArrayList<>();
        DataMatrix dMatrix = new DataMatrix();
        
        this.observations.forEach(obs -> {
            dMatrix.setParamDoubleValueForDate(obs.getTimeMeasured(), obs.getElementMeasurementTypeId(), obs.getValue());
        });
        List<Date> sortedMidnightTimestamps = dMatrix.getSortedDateAtMidnightKeys(this.timeZone);
        sortedMidnightTimestamps.forEach((Date midnightTimestamp) -> {
            Integer numberOfHoursWithDataForDay = 0;
            Integer wetHoursForDay = 0;
            XDate currentHour = new XDate(midnightTimestamp);
            Date tomorrow = currentHour.getTomorrow();
            
            while(currentHour.before(tomorrow))
            {
                Double TM = dMatrix.getParamDoubleValueForDate(currentHour, WeatherElements.TEMPERATURE_MEAN);
                Double UM = dMatrix.getParamDoubleValueForDate(currentHour, WeatherElements.RELATIVE_HUMIDITY_MEAN);
                Double BT = dMatrix.getParamDoubleValueForDate(currentHour, WeatherElements.LEAF_WETNESS_DURATION);
                Double RR = dMatrix.getParamDoubleValueForDate(currentHour, WeatherElements.PRECIPITATION);
                
                if(TM != null && (UM != null || BT != null || RR != null))
                {
                    numberOfHoursWithDataForDay++; // We add up the number of hours we actually have weather data for
                }

                if(TM != null && TM >= this.TEMPERATURE_THRESHOLD)
                {
                    wetHoursForDay += ((BT != null && BT > 30.0) || (RR != null && RR > 0.2) || (UM != null && UM > 88.0)) ? 1 : 0;
                }
                
                currentHour.add1Hour();
            }
            
            Result r = new ResultImpl();
            r.setValidTimeStart(midnightTimestamp);
            r.setValidTimeEnd(currentHour.get1HourBefore());
            if(numberOfHoursWithDataForDay < this.MINIMUM_HOURLY_VALUES)
            {
                r.setWarningStatus(Result.WARNING_STATUS_NO_WARNING_MISSING_DATA);
            }
            else
            {
                r.setValue(SeptoriaReferenceHumidityModel.MODEL_ID.toString(), DataMatrix.WET_HOUR_SUM, wetHoursForDay.toString());
                r.setWarningStatus(Result.WARNING_STATUS_NO_WARNING);
            }
            retVal.add(r);
        });
                
        return retVal;
    }

    @Override
    public ModelId getModelId() {
        return SeptoriaReferenceHumidityModel.MODEL_ID;
    }

    @Override
    public String getModelName() {
        return this.getModelName(DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelName(String language) {
        return "Septoria Reference Humidity Model";
    }

    @Override
    public String getLicense() {
        return "Copyright (c) 2019 NIBIO <http://www.nibio.no/>. \n" +
            "\n" +
            "This file is part of SeptoriaReferenceHumidityModel. \n" + 
            "SeptoriaReferenceHumidityModel is free software: you can redistribute it and/or modify \n" + 
            "it under the terms of the NIBIO Open Source License as published by \n" + 
            "NIBIO, either version 1 of the License, or (at your option) any \n" +
            "later version. \n" +
            "\n" +
            "SeptoriaReferenceHumidityModel is distributed in the hope that it will be useful, \n" +
            "but WITHOUT ANY WARRANTY; without even the implied warranty of \n" +
            "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \n" +
            "NIBIO Open Source License for more details. \n" +
            "" +
            "You should have received a copy of the NIBIO Open Source License \n" +
            "along with SeptoriaReferenceHumidityModel.  If not, see <http://www.nibio.no/licenses/>. \n";
    }

    @Override
    public String getCopyright() {
        return "(C) 2019 NIBIO";
    }

    @Override
    public String getModelDescription() {
        return this.getModelDescription(DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelDescription(String language) {
        return "<p>The reference humidity model was developed as a supplement to "
                + "the Humidity model. In this model 20 consecutive hours are "
                + "required to fulfil a risk period. One constraint in this "
                + "method is that you can have 19 consecutive risk hours  or,  "
                + "fx 14 hours with risk then one hour below the Rh threshold "
                + "and then maybe 14 hours again with risk hours. In one of "
                + "these situations, the model will indicate a risk. In the "
                + "reference model the definition of Humid hours was introduced. "
                + "The Rh threshold was avoided as humid hours do not need to be "
                + "consecutive. The running sum of humid hours across three days "
                + "indicate that the Septoria risk is higher than if you have "
                + "three days with humid conditions than two or one. The operation "
                + "of the model should include weather forecast data and should "
                + "run 6 days ahead from current day if you include a 7-day weather "
                + "forecast (60 hours from national met office and until 7 days from ECMWF)</p>" 
                
                +"<p>The model was tested against the Humidity model in a Danish "
                + "Septoria project funded by the GUDP. The threshold of 40 was "
                + "defined as high risk as this coincided with periods when the "
                + "humidity model recommended to apply fungicide (if not already protected). "
                + "The humidity model includes a decision model about when to spray, "
                + "protection periods ect. The reference model was used to quality "
                + "control the recommendations in a way that, if the reference humidity "
                + "hours were higher than 40 (no thresholds) then the user should "
                + "check the raw data for calculation of the Humidity model (threshold, "
                + "20 consecutive hours). If 2-3 periods of 17, 18, or 19 consecutive "
                + "hours appear, then one can consider to protect the crop based on the "
                + "reference model alone. </p>" 
                
                + "<p>The Humidity model is considered as a DSS with several components. "
                + "The reference humidity model is considered as a weather based submodel "
                + "for the risk of Septoria, Easy to map and calculate based on weather data alone.</p>";
    }

    @Override
    public String getWarningStatusInterpretation() {
        return this.getWarningStatusInterpretation(DEFAULT_LANGUAGE);
    }

    @Override
    public String getWarningStatusInterpretation(String language) {
        return "TODO";
    }

    @Override
    public String getModelUsage() {
        return this.getModelUsage(DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelUsage(String language) {
        return "Description of required input parameters:\n" +
"\n" +
"timeZone - What timezone the calculation is for. See this list of time zones: http://en.wikipedia.org/wiki/List_of_tz_database_time_zones (Column \"TZ\")\n" +
"\n" +
"observations - list of weather data. The following parameters are required:\n" +
"* BT - Leaf wetness (minutes/hour)\n" +
"* TM - Mean temp (Celcius)\n" +
"* UM - Relative humidity (%)\n" +
"* RR - Rainfall (mm)\n" +
"\n" +
"The weather data must be hourly data";
    }

    @Override
    public String getSampleConfig() {
        return "{\n" +
"	\"loginInfo\":{\n" +
"		\"username\":\"example\",\n" +
"		\"password\":\"example\"\n" +
"	},\n" +
"	\"modelId\":\"SEPAPIICOL\",\n" +
"	\"configParameters\":{\n" +
"		\"timeZone\":\"Europe/Oslo\",\n" +
"		\"observations\": [{\n" +
"			\"elementMeasurementTypeId\": \"TM\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T22:00:00Z\",\n" +
"			\"value\": 12.96\n" +
"		}, {\n" +
"			\"elementMeasurementTypeId\": \"BT\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T22:00:00Z\",\n" +
"			\"value\": 0.0\n" +
"		}, {\n" +
"			\"elementMeasurementTypeId\": \"UM\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T22:00:00Z\",\n" +
"			\"value\": 77.93\n" +
"		}, {\n" +
"			\"elementMeasurementTypeId\": \"RR\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T22:00:00Z\",\n" +
"			\"value\": 0.0\n" +
"		}, {\n" +
"			\"elementMeasurementTypeId\": \"TM\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T23:00:00Z\",\n" +
"			\"value\": 12.4\n" +
"		}, {\n" +
"			\"elementMeasurementTypeId\": \"BT\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T23:00:00Z\",\n" +
"			\"value\": 0.0\n" +
"		}, {\n" +
"			\"elementMeasurementTypeId\": \"UM\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T23:00:00Z\",\n" +
"			\"value\": 81.3\n" +
"		}, {\n" +
"			\"elementMeasurementTypeId\": \"RR\",\n" +
"			\"logIntervalId\": 1,\n" +
"			\"timeMeasured\": \"2019-04-30T23:00:00Z\",\n" +
"			\"value\": 0.0\n" +
"		}]\n" +
"	}\n" +
"}";
    }

    @Override
    public void setConfiguration(ModelConfiguration config) throws ConfigValidationException {
        ObjectMapper mapper = new ObjectMapper();
        this.timeZone = TimeZone.getTimeZone((String) config.getConfigParameter("timeZone"));
        this.observations = new ModelUtil().extractWeatherObservationList(config.getConfigParameter("observations"));
    }

}
